from __future__ import print_function, division
import argparse, gc, sys
import numpy as np
import pandas as pd
import gprim.dataset as gd
import ypy.memo as memo
import ypy.fs as fs
import ypy.pretty as pretty
import sldp.config as config

def main():
    parser = argparse.ArgumentParser()
    # optional arguments
    parser.add_argument('--spectrum-percent', type=float, default=95,
            help='Determines how many eigenvectors are kept in the truncated SVD. ' +\
                    'A value of x means that x percent of the eigenspectrum will be kept. '+\
                    'Default value: 95')
    parser.add_argument('--chroms', nargs='+', default=range(1,23), type=int,
            help='Space-delimited list of chromosomes to analyze. Default is 1..22')

    # configurable arguments
    parser.add_argument('--config', default=None,
            help='Path to a json file with values for other parameters. ' +\
                    'Values in this file will be overridden by any values passed ' +\
                    'explicitly via the command line.')
    parser.add_argument('--svd-stem', default=None,
            help='Path to directory in which output files will be stored. ' +\
                    'If not supplied, will be read from config file.')
    parser.add_argument('--bfile-chr', default=None,
            help='Path to plink bfile of reference panel to use, not including ' +\
                    'chromosome number. If not supplied, will be read from config file.')
    parser.add_argument('--print-snps', default=None,
            help='Path to set of potentially typed SNPs. If not supplied, will be read '+\
                    'from config file.')
    parser.add_argument('--ld-blocks', default=None,
            help='Path to UCSC bed file containing one bed interval per LD block. If '+\
                    'not supplied, will be read from config file.')

    print('=====')
    print(' '.join(sys.argv))
    print('=====')
    args = parser.parse_args()
    config.add_default_params(args)
    pretty.print_namespace(args)
    print('=====')

    # basic initialization
    mhc = [25684587, 35455756]
    refpanel = gd.Dataset(args.bfile_chr)
    fs.makedir_for_file(args.svd_stem)

    # read in ld blocks, remove MHC, read SNPs to print
    ldblocks = pd.read_csv(args.ld_blocks, delim_whitespace=True, header=None,
            names=['chr','start', 'end'])
    mhcblocks = (ldblocks.chr == 'chr6') & (ldblocks.end > mhc[0]) & (ldblocks.start < mhc[1])
    ldblocks = ldblocks[~mhcblocks]
    print(len(ldblocks), 'loci after removing MHC')
    print_snps = pd.read_csv(args.print_snps, header=None, names=['SNP'])
    print_snps['printsnp'] = True
    print(len(print_snps), 'print snps')

    for c in args.chroms:
        print('loading chr', c, 'of', args.chroms)
        # get refpanel snp metadata for this chromosome
        snps = refpanel.bim_df(c)
        snps = pd.merge(snps, print_snps, on='SNP', how='left')
        snps.printsnp.fillna(False, inplace=True)
        print(len(snps), 'snps in refpanel', len(snps.columns), 'columns, including metadata')

        # iterate through ld blocks and process each one
        for ldblock, X, meta, _ in refpanel.block_data(ldblocks, c, meta=snps):
            if meta.printsnp.sum() == 0:
                print('no print snps found in this block')
                continue

            # restrict X to have only potentially typed SNPs along one axis
            mask = meta.printsnp.values
            X_ = X[:,mask]

            def bestsvd(A):
                try:
                    U_, svs_, _ = np.linalg.svd(A.T); svs_ = svs_**2 / A.shape[0]
                except np.linalg.linalg.LinAlgError:
                    print('\t\tresorting to svd of XTX')
                    U_, svs_, _ = np.linalg.svd(A.T.dot(A)); svs_ = svs_ / A.shape[0]
                return U_, svs_

            # compute the (right-hand) SVD of R
            print('\tcomputing SVD of R_print')
            U_, svs_ = bestsvd(X_)
            k = np.argmax(np.cumsum(svs_)/svs_.sum() >= args.spectrum_percent / 100.)
            print('\treduced rank of', k, 'out of', meta.printsnp.sum(), 'printed snps')
            np.savez('{}{}.R'.format(args.svd_stem, ldblock.name), U=U_[:,:k], svs=svs_[:k])

            # compute the SVD of R2
            print('\tcomputing R2_print')
            R2 = X_.T.dot(X.dot(X.T)).dot(X_) / X.shape[0]**2
            print('\tcomputing SVD of R2_print')
            R2_U, R2_svs, _ = np.linalg.svd(R2)
            k = np.argmax(np.cumsum(R2_svs)/R2_svs.sum() >= args.spectrum_percent / 100.)
            print('\treduced rank of', k, 'out of', meta.printsnp.sum(), 'printed snps')
            np.savez('{}{}.R2'.format(args.svd_stem, ldblock.name),
                    U=R2_U[:,:k], svs=R2_svs[:k])

        del snps; memo.reset(); gc.collect()
    print('done')


if __name__ == '__main__':
    main()



from __future__ import print_function, division
import argparse, sys
import sldp.config as config
import ypy.pretty as pretty

def do(args):
    print('initializing...')
    import gzip, gc, time
    import numpy as np
    import pandas as pd
    import gprim.annotation as ga
    import gprim.dataset as gd
    import ypy.memo as memo

    # basic initialization
    mhc_bp = [25684587, 35455756]
    refpanel = gd.Dataset(args.bfile_chr)
    annots = [ga.Annotation(annot) for annot in args.sannot_chr]

    # read in ld blocks, remove MHC, read SNPs to print
    ldblocks = pd.read_csv(args.ld_blocks, delim_whitespace=True, header=None,
            names=['chr','start', 'end'])
    mhcblocks = (ldblocks.chr == 'chr6') & \
            (ldblocks.end > mhc_bp[0]) & \
            (ldblocks.start < mhc_bp[1])
    ldblocks = ldblocks[~mhcblocks]
    print(len(ldblocks), 'loci after removing MHC')
    print_snps = pd.read_csv(args.print_snps, header=None, names=['SNP'])
    print_snps['printsnp'] = True
    print(len(print_snps), 'print snps')

    # process annotations
    for annot in annots:
        t0 = time.time()
        for c in args.chroms:
            print(time.time()-t0, ': loading chr', c, 'of', args.chroms)
            # get refpanel snp metadata for this chromosome
            snps = refpanel.bim_df(c)
            snps = ga.smart_merge(snps, refpanel.frq_df(c)[['SNP','MAF']])
            print(len(snps), 'snps in refpanel',
                    len(snps.columns), 'columns, including metadata')

            # read in annotation
            print('reading annot', annot.filestem())
            names = annot.names(c) # names of annotations
            namesR = [n+'.R' for n in names] # names of results
            a = annot.sannot_df(c)
            if 'SNP' in a.columns:
                print('not a thinannot => doing full reconciliation of snps and allele coding')
                snps = ga.reconciled_to(snps, a, names, missing_val=0)
            else:
                print('detected thinannot, so assuming that annotation is synched to refpanel')
                snps = pd.concat([snps, a[names]], axis=1)

            # add information on which snps to print
            print('merging in print_snps')
            snps = pd.merge(snps, print_snps, how='left', on='SNP')
            snps.printsnp.fillna(False, inplace=True)
            snps.printsnp.astype(bool)

            # put on per-normalized-genotype scale
            if args.alpha != -1:
                print('scaling by maf according to alpha=', args.alpha)
                snps[names] = snps[names].values*\
                        np.power(2*snps.MAF.values*(1-snps.MAF.values),
                                (1.+args.alpha)/2)[:,None]

            # make room for RV and make sure annotation values are treated as floats
            snps = pd.concat(
                    [snps, pd.DataFrame(np.zeros(snps[names].shape), columns=namesR)],
                    axis=1)
            snps[names] = snps[names].astype(float)

            # compute simple statistics about annotation
            print('computing basic statistics and writing')
            info = pd.DataFrame(
                    columns=['M', 'M_5_50', 'sqnorm', 'sqnorm_5_50', 'supp', 'supp_5_50'])
            info['name'] = names
            info.set_index('name', inplace=True)
            info['M'] = len(snps)
            info['sqnorm'] = np.linalg.norm(snps[names], axis=0)**2
            info['supp'] = np.linalg.norm(snps[names], ord=0, axis=0)
            M_5_50 = (snps.MAF >= 0.05).values
            info['M_5_50'] = M_5_50.sum()
            info['sqnorm_5_50'] = np.linalg.norm(snps.loc[M_5_50, names], axis=0)**2
            info['supp_5_50'] = np.linalg.norm(snps.loc[M_5_50, names], ord=0, axis=0)
            info.to_csv(annot.info_filename(c), sep='\t')

            # process ldblocks one by one
            for ldblock, X, meta, ind in refpanel.block_data(ldblocks, c, meta=snps):
                if meta.printsnp.sum() == 0:
                    print('no print-snps in this block')
                    continue
                print(meta.printsnp.sum(), 'print-snps')
                if (meta[names] == 0).values.all():
                    print('annotations are all 0 in this block')
                    snps.loc[ind, namesR] = 0
                else:
                    mask = meta.printsnp.values
                    V = meta[names].values
                    XV = X.dot(V)
                    snps.loc[ind[mask], namesR] = \
                            X[:,mask].T.dot(XV[:,-len(names):]) / X.shape[0]

            # write
            print('writing output')
            with gzip.open(annot.RV_filename(c), 'w') as f:
                snps.loc[snps.printsnp,['SNP','A1','A2']+names+namesR].to_csv(
                        f, index=False, sep='\t')

            del snps; memo.reset(); gc.collect()

    print('done')

def main():
    parser = argparse.ArgumentParser()
    # required arguments
    parser.add_argument('--sannot-chr', nargs='+', required=True,
            help='Multiple (space-delimited) paths to sannot.gz files, not including '+\
                    'chromosome')

    # optional arguments
    parser.add_argument('--alpha', type=float, default=-1,
        help='scale annotation values by sqrt(2*maf(1-maf))^{alpha+1}. '+\
                '-1 means assume annotation values are already per-normalized-genotype, '+\
                '0 means assume they were per allele. Default is -1.')
    parser.add_argument('--chroms', nargs='+', default=range(1,23), type=int,
            help='Space-delimited list of chromosomes to analyze. Default is 1..22')

    # configurable arguments
    parser.add_argument('--config', default=None,
            help='Path to a json file with values for other parameters. ' +\
                    'Values in this file will be overridden by any values passed ' +\
                    'explicitly via the command line.')
    parser.add_argument('--bfile-chr', default=None,
            help='Path to plink bfile of reference panel to use, not including ' +\
                    'chromosome number. If not supplied, will be read from config file.')
    parser.add_argument('--print-snps', default=None,
            help='Path to set of potentially typed SNPs. If not supplied, will be read '+\
                    'from config file.')
    parser.add_argument('--ld-blocks', default=None,
            help='Path to UCSC bed file containing one bed interval per LD block. If '+\
                    'not supplied, will be read from config file.')

    print('=====')
    print(' '.join(sys.argv))
    print('=====')
    args = parser.parse_args()
    config.add_default_params(args)
    pretty.print_namespace(args)
    print('=====')

    do(args)

if __name__ == "__main__":
    main()


from __future__ import print_function, division
import argparse, sys
import sldp.config as config
import ypy.pretty as pretty

def main():
    parser = argparse.ArgumentParser()
    # required arguments
    parser.add_argument('--outfile-stem', required=True,
            help='Path to an output file stem.')
    pheno = parser.add_mutually_exclusive_group(required=True)
    pheno.add_argument('--pss-chr', default=None,
            help='Path to .pss.gz file, without chromosome number or .pss.gz extension. '+\
                    'This is the phenotype that SLDP will analyze.')
    pheno.add_argument('--sumstats-stem', default=None,
            help='Path to a .sumstats.gz file, not including ".sumstats.gz" extension. '+\
                    'SLDP will process this into a set of .pss.gz files before running.')
    parser.add_argument('--sannot-chr', nargs='+', required=True,
            help='One or more (space-delimited) paths to gzipped annot files, without '+\
                    'chromosome number or .sannot.gz/.RV.gz extension. These are the '+\
                    'annotations that SLDP will analyze against the phenotype.')

    # optional arguments
    parser.add_argument('--verbose-thresh', default=0., type=float,
            help='Print additional information about each association studied with a '+\
                    'p-value below this number. (Default is 0.) This includes: '+\
                    'the covariance in each independent block of genome (.chunks files), '+\
                    'and the coefficients required to residualize any background '+\
                        'annotations out of the other annotations being analyzed.')
    parser.add_argument('-fastp', default=False, action='store_true',
            help='Estimate p-values fast (without permutation)')
    parser.add_argument('-bothp', default=False, action='store_true',
            help='Print both fastp p-values (as p_fast) and normal p-values. '+\
                    'Takes precedence over fastp')
    parser.add_argument('--tell-me-stories', default=0.,
            help='!!Experimental!! For associations with a p-value less than this number, '+\
                    'print information about loci that may be promising to study. '+\
                    'This will produce plots of (potentially overlapping) loci where the '+\
                    'signed LD profile is highly correlated with the GWAS signal in a '+\
                    'direction consistent with the global effect. '+\
                    'Default value is 0.')
    parser.add_argument('--story-corr-thresh', default=0.8, type=float,
            help='The threshold to use for correlation between Rv and alphahat in order '+\
                    'for a locus to be considered worthy of a story')
    parser.add_argument('-more-stats', default=False, action='store_true',
            help='Print additional statistis about q in results file')
    parser.add_argument('--T', type=int, default=1000000,
            help='number of times to sign flip for empirical p-values. Default is 10^6.')
    parser.add_argument('--jk-blocks', type=int, default=300,
            help='Number of jackknife blocks to use. Default is 300.')
    parser.add_argument('--weights', default='Winv_ahat_h',
            help='which set of regression weights to use. Default is Winv_ahat_h, '+\
                    'corresponding to weights described in Reshef et al. 2017.')
    parser.add_argument('--seed', default=None, type=int,
            help='Seed random number generator to a certain value. Off by default.')
    parser.add_argument('--stat', default='sum',
            help='*experimental* Which statistic to use for hypothesis testing. Options '+\
                    'are: sum, medrank, or thresh.')
    parser.add_argument('--chi2-thresh', default=0, type=float,
            help='only use SNPs with a chi2 above this number for the regression')
    parser.add_argument('--chroms', nargs='+', default=range(1,23), type=int,
            help='Space-delimited list of chromosomes to analyze. Default is 1..22')

    # configurable arguments
    parser.add_argument('--config', default=None,
            help='Path to a json file with values for other parameters. ' +\
                    'Values in this file will be overridden by any values passed ' +\
                    'explicitly via the command line.')
    parser.add_argument('--background-sannot-chr', nargs='+', default=[],
            help='One or more (space-delimited) paths to gzipped annot files, without '+\
                    'chromosome number or .sannot.gz extension. These are the annotations '+\
                    'that SLDP will control for.')
    parser.add_argument('--svd-stem', default=None,
            help='Path to directory in which output files will be stored. ' +\
                    'If not supplied, will be read from config file.')
    parser.add_argument('--bfile-reg-chr', default=None,
            help='Path to plink bfile of reference panel to use, not including ' +\
                    'chromosome number. This bfile should contain only regression SNPs '+\
                    '(as opposed to, e.g., all potentially causal SNPs). '+\
                    'If not supplied, will be read from config file.')
    parser.add_argument('--ld-blocks', default=None,
            help='Path to UCSC bed file containing one bed interval per LD block. If '+\
                    'not supplied, will be read from config file.')


    print('=====')
    print(' '.join(sys.argv))
    print('=====')
    args = parser.parse_args()
    config.add_default_params(args)
    pretty.print_namespace(args)
    print('=====')

    preprocess_sumstats(args)
    preprocess_sannots(args)

    print('initializing...')
    import os, time
    import numpy as np
    import pandas as pd
    import scipy.stats as st
    import gprim.annotation as ga
    import gprim.dataset as gd
    import ypy.memo as memo
    import sldp.weights as weights
    import sldp.chunkstats as cs
    import sldp.storyteller as storyteller

    # basic initialization
    mhc_bp = [25684587, 35455756]
    refpanel = gd.Dataset(args.bfile_reg_chr)
    pheno_name = args.pss_chr.split('/')[-2].replace('.KG3.95','')
    if args.seed is not None:
        np.random.seed(args.seed)
        print('random seed:', args.seed)

    # read in names of background annotations and marginal annotations
    annots = [ga.Annotation(annot) for annot in args.sannot_chr]
    marginal_names = sum([a.names(22, RV=True) for a in annots], [])
    marginal_names = [n for n in marginal_names if '.R' in n]
    marginal_infos = pd.concat([a.info_df(args.chroms) for a in annots], axis=0)
    backgroundannots = [ga.Annotation(annot) for annot in args.background_sannot_chr]
    background_names = sum([a.names(22, True) for a in backgroundannots], [])
    background_names = [n for n in background_names if '.R' in n]
    print('background annotations:', background_names)
    print('marginal annotations:', marginal_names)
    if len(set(background_names) & set(marginal_names)) > 0:
        raise ValueError('the background annotation names and the marginal annotation '+\
                'names must be disjoint sets')

    # read in ldblocks and remove ones that overlap mhc
    ldblocks = pd.read_csv(args.ld_blocks, delim_whitespace=True, header=None,
            names=['chr','start', 'end'])
    mhcblocks = (ldblocks.chr == 'chr6') & \
            (ldblocks.end > mhc_bp[0]) & \
            (ldblocks.start < mhc_bp[1])
    ldblocks = ldblocks[~mhcblocks]

    # read information about sumstats
    sumstats_info = pd.read_csv(args.pss_chr+'info', sep='\t')
    sigma2g = sumstats_info.loc[0].sigma2g
    h2g = sumstats_info.loc[0].h2g

    # read in sumstats and annots, and compute numerator and denominator of regression for
    # each ldblock. These will later be processed and aggregated
    numerators = dict(); denominators = dict()
    t0 = time.time()
    for c in args.chroms:
        print(time.time()-t0, ': loading chr', c, 'of', args.chroms)

        # get refpanel snp metadata 
        snps = refpanel.bim_df(c)
        print(len(snps), 'snps in refpanel', len(snps.columns), 'columns, including metadata')

        # read sumstats
        print('reading sumstats')
        ss = pd.read_csv(args.pss_chr+str(c)+'.pss.gz', sep='\t')
        print(np.isnan(ss[args.weights]).sum(), 'sumstats nans out of', len(ss))
        snps['Winv_ahat'] = ss[args.weights]
        snps['N'] = ss.N
        snps['typed'] = snps.Winv_ahat.notnull()
        if args.chi2_thresh > 0:
            print('applying chi2 threshold of', args.chi2_thresh)
            snps.typed &= (ss.Winv_ahat_I**2 * ss.N > args.chi2_thresh)
            print(snps.typed.sum(), 'typed snps left')

        # read annotations
        print('reading annotations')
        for annot in backgroundannots+annots:
            mynames = [n for n in annot.names(22, RV=True) if '.R' in n] #names of annotations
            print(time.time()-t0, ': reading annot', annot.filestem())
            print('adding', mynames)
            snps = pd.concat([snps, annot.RV_df(c)[mynames]], axis=1)
            if (~np.isfinite(snps[mynames].values)).sum() > 0:
                raise ValueError('There should be no nans in the postprocessed annotation. '+\
                        'But there are '+str((~np.isfinite(snps[mynames].values)).sum()))

        # make sure things are in the order we think they are
        if (np.array(background_names+marginal_names) !=
                snps.columns.values[-len(background_names+marginal_names):]).any():
            raise ValueError('Merged annotations are not in the right order')

        # perform computations
        for ldblock, _, meta, ind in refpanel.block_data(
                ldblocks, c, meta=snps, genos=False, verbose=0):
            if meta.typed.sum() == 0 or \
                    not os.path.exists(args.svd_stem+str(ldblock.name)+'.R.npz'):
                # no typed snps/hm3 snps in this block. set num snps to 0
                ldblocks.loc[ldblock.name, 'M_H'] = 0
                continue
            if (meta[background_names+marginal_names] == 0).values.all():
                # annotations are all 0 in this block
                ldblocks.loc[ldblock.name, 'M_H'] = 0
                continue
            # record the number of typed snps in this block, and start- and end- snp indices
            ldblocks.loc[ldblock.name, 'M_H'] = meta.typed.sum()
            ldblocks.loc[ldblock.name, 'snpind_begin'] = min(meta.index) # for verbose
            ldblocks.loc[ldblock.name, 'snpind_end'] = max(meta.index)+1 # for verbose

            # load regression weights and prepare for regression computation
            meta_t = meta[meta.typed.values]
            N = meta_t.N.mean()
            if args.weights == 'Winv_ahat_h' or args.weights == 'Winv_ahat_hlN':
                R = np.load(args.svd_stem+str(ldblock.name)+'.R.npz')
                R2 = None
                if len(R['U']) != len(meta):
                    raise ValueError('regression wgts dimension must match regression snps')
            elif args.weights == 'Winv_ahat_h2' or args.weights == 'Winv_ahat':
                R = np.load(args.svd_stem+str(ldblock.name)+'.R.npz')
                R2 = np.load(args.svd_stem+str(ldblock.name)+'.R2.npz')
                if len(R['U']) != len(meta) or len(R2['U']) != len(meta):
                    raise ValueError('regression wgts dimension must match regression snps')
            else:
                R = None; R2 = None

            # multiply ahat by the weights
            Winv_RV = weights.invert_weights(
                    R, R2, sigma2g, N, meta[background_names+marginal_names].values,
                    typed=meta.typed.values, mode=args.weights)

            numerators[ldblock.name] = \
                (meta_t[background_names+marginal_names].T.dot(
                    meta_t.Winv_ahat)).values/1e6
            denominators[ldblock.name] = meta_t[background_names+marginal_names].T.dot(
                    Winv_RV[meta.typed.values]).values/1e6
        memo.reset()

    # get data for jackknifing
    print('jackknifing')
    chunk_nums, chunk_denoms, loo_nums, loo_denoms, chunkinfo = cs.collapse_to_chunks(
            ldblocks,
            numerators,
            denominators,
            args.jk_blocks)

    # compute final results
    global q, results
    results = pd.DataFrame()
    for i, name in enumerate(marginal_names):
        print(i, name)
        # metadata about v
        sqnorm = marginal_infos.loc[name[:-2], 'sqnorm']
        supp = marginal_infos.loc[name[:-2], 'supp']

        # estimate mu and initialize results row
        k = marginal_names.index(name)
        mu = cs.get_est(sum(chunk_nums), sum(chunk_denoms), k, len(background_names))
        newrow_results = {
            'pheno':pheno_name,
            'annot':name}
        results = pd.concat([results, pd.DataFrame([newrow_results])], ignore_index=True)

        # compute q
        q, r, mux, muy = cs.residualize(chunk_nums, chunk_denoms, len(background_names), k)


        # p-values
        if args.bothp or not args.fastp:
            p_emp, z_emp = cs.signflip(q, args.T, printmem=True, mode=args.stat)
            results.loc[i,'z'] = z_emp
            results.loc[i,'p'] = p_emp
        if args.bothp or args.fastp:
            z_fast = np.sum(q)/np.linalg.norm(q)
            p_fast = st.chi2.sf(z_fast**2, 1)
            results.loc[i, 'z_fast'] = z_fast
            results.loc[i, 'p_fast'] = p_fast
        if args.fastp and not args.bothp: # if only fastp, rename the output columns
            results.loc[i,'p'] = results.loc[i,'p_fast']
            results.loc[i,'z'] = results.loc[i,'z_fast']
            results.drop(['pfast','zfast'], axis=1, inplace=True)

        # print verbose information if required
        if results.loc[i].p < args.verbose_thresh:
            fname = args.outfile_stem+'.'+pheno_name+'.'+name
            print('writing verbose results to', fname)
            chunkinfo['q'] = q
            chunkinfo['r'] = r
            chunkinfo.to_csv(fname+'.chunks', sep='\t', index=False)
            coeffs = pd.DataFrame()
            coeffs['annot'] = background_names
            coeffs['mux'] = mux
            coeffs['muy'] = muy
            coeffs.to_csv(fname+'.coeffs', sep='\t', index=False)

        # nominate interesting loci if desired
        if results.loc[i].p < args.tell_me_stories:
            storyteller.write(args.outfile_stem+'.'+name+'.loci',
                    args, name, background_names, mux, muy, results.loc[i].z,
                    corr_thresh=args.story_corr_thresh)

        # add more output if desired
        if args.more_stats:
            results.loc[i,'qkurtosis'] = st.kurtosis(q)
            results.loc[i,'qstd'] = np.std(q)
            results.loc[i,'p_jk'] = st.chi2.sf((mu/se)**2,1)
            results.loc[i,'sqnorm'] = sqnorm

        # jackknife
        se = cs.jackknife_se(mu, loo_nums, loo_denoms, k, len(background_names))
        results.loc[i,'mu'] = mu
        results.loc[i,'se(mu)'] = se

        # add estimates of rf, h2v, and other associated quantities
        M = marginal_infos.loc[name[:-2],'M']
        results.loc[i,'h2g'] = h2g
        results.loc[i,'rf'] = mu * np.sqrt(
                sqnorm / h2g)
        results.loc[i,'h2v/h2g'] = results.loc[i].rf**2 - \
                results.loc[i,'se(mu)']**2 * sqnorm / (M*sigma2g)
        results.loc[i,'h2v'] = results.loc[i,'h2v/h2g'] * h2g
        results.loc[i,'supp(v)/M'] = supp/M
        results.to_csv(args.outfile_stem + '.gwresults', sep='\t', index=False, na_rep='nan')

    print(results)
    print('writing results to', args.outfile_stem + '.gwresults')
    results.to_csv(args.outfile_stem + '.gwresults', sep='\t', index=False, na_rep='nan')
    print('done')

# preprocess any sumstats that need preprocessing
def preprocess_sumstats(args):
    import os
    if args.pss_chr is None:
        unprocessed_chroms = [
                c for c in args.chroms
                if not os.path.exists(
                    args.sumstats_stem + '.' + args.refpanel_name + '/'+str(c)+'.pss.gz')
                ]
        if len(unprocessed_chroms) > 0:
            print('Preprocessing', args.sumstats_stem+'.sumstats.gz... at', unprocessed_chroms)
            if args.config is None:
                raise ValueError('automatic pre-processing of a sumstats file requires '+\
                        'specification of a config file; otherwise I dont know what '+\
                        'parameters to use. If you want, you can preprocess the sumstats '+\
                        'without a config file by running preprocesspheno manually')
            print('Using config file', args.config, 'and default options')

            # run the command
            import sldp.preprocesspheno, copy
            args_ = copy.copy(args)
            args_.no_M_5_50 = False
            args_.set_h2g = None
            args_.chroms = unprocessed_chroms
            sldp.preprocesspheno.main(args_)

        # modify args to reflect existing of pss-chr files
        args.pss_chr = args.sumstats_stem + '.' + args.refpanel_name + '/'
        args.sumstats_stem = None
        print('== finished preprocessing sumstats ==')

# preprocess any annotations that need preprocessing
def preprocess_sannots(args):
    import os

    for sannot in args.sannot_chr:
        unprocessed_chroms = [
                c for c in args.chroms
                if not (os.path.exists(sannot + str(c) + '.RV.gz') and
                    os.path.exists(sannot + str(c) + '.info'))
                ]
        if len(unprocessed_chroms) > 0:
            print('Preprocessing', sannot, 'at chromosomes', unprocessed_chroms)
            if args.config is None:
                raise ValueError('automatic pre-processing of an annotation '+\
                        'requires specification of a config file; otherwise I dont know what '+\
                        'parameters to use. If you want, you can preprocess the annotation '+\
                        'without a config file by running preprocessannot manually')
            print('Using config file', args.config, 'and default options')

            # run preprocessing command
            import sldp.preprocessannot, copy
            args_ = copy.copy(args)
            args_.alpha = -1
            args_.chroms = unprocessed_chroms
            sldp.preprocessannot.main(args_)

            print('== finished preprocessing annotation', sanno)


if __name__ == '__main__':
    main()



from __future__ import print_function, division
import argparse, sys
import sldp.config as config
import ypy.pretty as pretty

def do(args):
    print('initializing...')
    import gzip, os, gc, time
    import numpy as np
    import pandas as pd
    import gprim.annotation as ga
    import gprim.dataset as gd
    import ypy.fs as fs
    import ypy.memo as memo
    import sldp.weights as weights

    # read in refpanel, ld blocks, and svd snps
    refpanel = gd.Dataset(args.bfile_chr)
    ldblocks = pd.read_csv(args.ld_blocks, delim_whitespace=True, header=None,
            names=['chr','start', 'end'])
    print_snps = pd.read_csv(args.print_snps, header=None, names=['SNP'])
    print_snps['printsnp'] = True
    print(len(print_snps), 'svd snps')

    # read sumstats
    print('reading sumstats', args.sumstats_stem)
    ss = pd.read_csv(args.sumstats_stem+'.sumstats.gz', sep='\t')
    ss = ss[ss.Z.notnull() & ss.N.notnull()]
    print('{} snps, {}-{} individuals (avg: {})'.format(
        len(ss), np.min(ss.N), np.max(ss.N), np.mean(ss.N)))
    ss = pd.merge(ss, print_snps[['SNP']], on='SNP', how='inner')
    print(len(ss), 'snps typed')

    # read ld scores
    print('reading in ld scores')
    ld = pd.concat([pd.read_csv(args.ldscores_chr+str(c)+'.l2.ldscore.gz',
                        delim_whitespace=True)
                    for c in range(1,23)], axis=0)
    if args.no_M_5_50:
        M = sum([int(open(args.ldscores_chr+str(c)+'.l2.M').next())
                        for c in range(1,23)])
    else:
        M = sum([int(open(args.ldscores_chr+str(c)+'.l2.M_5_50').next())
                        for c in range(1,23)])
    print(len(ld), 'snps with ld scores')
    ssld = pd.merge(ss, ld, on='SNP', how='left')
    print(len(ssld), 'hm3 snps with sumstats after merge.')

    # estimate heritability using aggregate estimator
    def esth2g(ssld):
        meanchi2 = (ssld.Z**2).mean()
        meanNl2 = (ssld.N*ssld.L2).mean()
        sigma2g = (meanchi2 - 1)/meanNl2
        h2g = sigma2g * M
        K = M/meanNl2 # h2g = K (meanchi2 - 1)
        return h2g, sigma2g, meanchi2, K
    h2g, sigma2g, meanchi2, K = esth2g(ssld)
    h2g = max(h2g, 0.03) #0.03 is an arbitrarily chosen minimum
    print('mean chi2:', meanchi2)
    print('h2g estimated at:', h2g, 'sigma2g =', sigma2g)
    if args.set_h2g:
        print('scaling Z-scores to achieve h2g of', args.set_h2g)
        norm = meanchi2 / (1 + args.set_h2g/ K)
        print('dividing all z-scores by', np.sqrt(norm))
        ssld.Z /= np.sqrt(norm)
        h2g, sigma2g, _, _ = esth2g(ssld)
        print('h2g is now', h2g)

    # write h2g results to file
    dirname = args.sumstats_stem + '.' + args.refpanel_name
    fs.makedir(dirname)
    if 1 in args.chroms:
        print('writing info file')
        info = pd.DataFrame(); info=info.append(
                {'pheno':args.sumstats_stem.split('/')[-1],
                    'h2g':h2g,
                    'sigma2g':sigma2g,
                    'Nbar':ss.N.mean()},ignore_index=True)
        info.to_csv(dirname+'/info', sep='\t', index=False)

    # preprocess ld blocks
    t0 = time.time()
    for c in args.chroms:
        print(time.time()-t0, ': loading chr', c, 'of', args.chroms)
        # get refpanel snp metadata for this chromosome
        snps = refpanel.bim_df(c)
        snps = pd.merge(snps, print_snps, on='SNP', how='left')
        snps.printsnp.fillna(False, inplace=True)
        print(len(snps), 'snps in refpanel', len(snps.columns), 'columns, including metadata')

        # merge annot and sumstats
        print('reconciling')
        snps = ga.reconciled_to(snps, ss, ['Z'],
                othercolnames=['N'], missing_val=np.nan)
        snps['typed'] = snps.Z.notnull()
        snps['ahat'] = snps.Z / np.sqrt(snps.N)

        # initialize result dataframe
        # I = no weights
        # h = heuristic weights, using R_o
        snps['Winv_ahat_I'] = np.nan # = W_o^{-1} ahat_o
        snps['R_Winv_ahat_I'] = np.nan # = R_{*o} W_o^{-1} ahat_o
        snps['Winv_ahat_h'] = np.nan # = W_o^{-1} ahat_o
        snps['R_Winv_ahat_h'] = np.nan # = R_{*o} W_o^{-1} ahat_o

        # restrict to ld blocks in this chr and process them in chunks
        for ldblock, X, meta, ind in refpanel.block_data(ldblocks, c, meta=snps):
            if meta.printsnp.sum() == 0 or \
                    not os.path.exists(args.svd_stem+str(ldblock.name)+'.R.npz'):
                print('no svd snps found in this block')
                continue
            print(meta.printsnp.sum(), 'svd snps', meta.typed.sum(), 'typed snps')
            if meta.typed.sum() == 0:
                print('no typed snps found in this block')
                snps.loc[ind, [
                    'R_Winv_ahat_I',
                    'R_Winv_ahat_h'
                    ]] = 0
                continue
            R = np.load(args.svd_stem+str(ldblock.name)+'.R.npz')
            R2 = np.load(args.svd_stem+str(ldblock.name)+'.R2.npz')
            N = meta[meta.typed.values].N.mean()
            meta_svd = meta[meta.printsnp.values]

            # multiply ahat by the weights
            x_I = snps.loc[ind[meta.printsnp],'Winv_ahat_I'] = weights.invert_weights(
                    R, R2, sigma2g, N, meta_svd.ahat.values, mode='Winv_ahat_I')
            x_h = snps.loc[ind[meta.printsnp],'Winv_ahat_h'] = weights.invert_weights(
                    R, R2, sigma2g, N, meta_svd.ahat.values, mode='Winv_ahat_h')

        print('writing processed sumstats')
        with gzip.open('{}/{}.pss.gz'.format(dirname, c), 'w') as f:
            snps.loc[snps.printsnp,['N',
                'Winv_ahat_I',
                'Winv_ahat_h'
                ]].to_csv(
                    f, index=False, sep='\t')

        del snps; memo.reset(); gc.collect()

    print('done')

def main():
    parser = argparse.ArgumentParser()
    # required arguments
    parser.add_argument('--sumstats-stem', required=True,
            help='path to sumstats.gz files, not including ".sumstats.gz" extension')

    # optional arguments
    parser.add_argument('--refpanel-name', default='KG3.95',
            help='suffix added to the directory created for storing output. '+\
                    'Default is KG3.95, corresponding to 1KG Phase 3 reference panel '+\
                    'processed with default parameters by preprocessrefpanel.py.')
    parser.add_argument('-no-M-5-50', default=False, action='store_true',
            help='Dont filter to SNPs with MAF >= 0.05 when estimating heritabilities')
    parser.add_argument('--set-h2g', default=None, type=float,
            help='Scale Z-scores to achieve this approximate heritability')
    parser.add_argument('--chroms', nargs='+', default=range(1,23), type=int,
            help='Space-delimited list of chromosomes to analyze. Default is 1..22')

    # configurable arguments
    parser.add_argument('--config', default=None,
            help='Path to a json file with values for other parameters. ' +\
                    'Values in this file will be overridden by any values passed ' +\
                    'explicitly via the command line.')
    parser.add_argument('--bfile-chr', default=None,
            help='Path to plink bfile of reference panel to use, not including ' +\
                    'chromosome number. If not supplied, will be read from config file.')
    parser.add_argument('--svd-stem', default=None,
            help='Path to directory containing truncated svds of reference panel, by LD '+\
                    'block, as produced by preprocessrefpanel.py. If not supplied, will be '+\
                    'read from config file.')
    parser.add_argument('--print-snps', default=None,
            help='Path to set of potentially typed SNPs. If not supplied, will be read '+\
                    'from config file.')
    parser.add_argument('--ldscores-chr', default=None,
            help='Path to LD scores at a smallish set of SNPs (~1M). LD should be computed '+\
                    'to all potentially causal snps. Used for heritability estimation. '+\
                    'If not supplied, will be read from config file.')
    parser.add_argument('--ld-blocks', default=None,
            help='Path to UCSC bed file containing one bed interval per LD block. If '+\
                    'not supplied, will be read from config file.')

    print('=====')
    print(' '.join(sys.argv))
    print('=====')
    args = parser.parse_args()
    config.add_default_params(args)
    pretty.print_namespace(args)
    print('=====')

    do(args)

if __name__ == '__main__':
    main()


from __future__ import print_function, division
import numpy as np
import pandas as pd
import gprim.annotation as ga
import gprim.dataset as gd
import ypy.fs as fs
import matplotlib.pyplot as plt

# find windows with genome-wide significant SNPs that are consistent with the global signal
def write(folder, args, name, background_names, mux, muy, z, corr_thresh=0.8):
    print('STORYTELLING for ', name, 'z=', z)
    refpanel = gd.Dataset(args.bfile_reg_chr)
    annot = [ga.Annotation(a) for a in args.sannot_chr
            if name in ga.Annotation(a).names(22, RV=True)][0]

    backgroundannots = [ga.Annotation(a) for a in args.background_sannot_chr]
    print('focal annotation columns:', annot.names(22, True))
    print('background annotations:', background_names)

    # get refpanel snp metadata 
    print('re-reading snps')
    snps = pd.concat([refpanel.bim_df(c) for c in args.chroms], axis=0)

    # read sumstats
    print('re-reading sumstats')
    ss = pd.concat([
        pd.read_csv(args.pss_chr+str(c)+'.pss.gz', sep='\t', usecols=['N','Winv_ahat_I'])
        for c in args.chroms])
    snps['ahat'] = ss['Winv_ahat_I']
    snps['N'] = ss['N']
    del ss

    # read annotations
    print('re-reading background annotations')
    for a in backgroundannots:
        mynames = [n for n in a.names(22, RV=True) if '.R' in n] #names of annotations
        snps = pd.concat([snps,
            pd.concat([a.RV_df(c)[mynames] for c in args.chroms], axis=0)
            ], axis=1)

    print('reading focal annotation')
    snps = pd.concat([snps,
        pd.concat([annot.RV_df(c)[name] for c in args.chroms], axis=0)
        ], axis=1)

    print('residualizing background out of focal')
    A = snps[background_names]
    snps['chi2'] = snps.N * snps.ahat**2
    snps['Rv'] = snps[name]
    snps['ahat_resid'] = snps.ahat - A.values.dot(muy)
    snps['Rv_resid'] = snps.Rv - A.values.dot(mux)
    snps['typed'] = snps.ahat_resid.notnull()
    snps = snps[snps.typed].reset_index(drop=True)
    snps['significant'] = snps.chi2 > 29.716785
    print(snps.significant.sum(), 'genome-wide significant SNPs')

    print('searching for good windows')
    # get endpoints of windows
    stride = 20
    windowsize_in_strides = 5
    windowsize = stride * windowsize_in_strides

    # find all starting points of windows containing GWAS-sig SNPs
    starts = np.concatenate([
                [ int(i/stride)*stride - k*stride
                    for k in range(0, windowsize_in_strides)]
                for i in np.where(snps.significant)[0]])
    starts = np.array(sorted(list(set(starts))))
    # compute corresponding endpoints
    ends = starts + windowsize

    # truncate any windows that extend past the ends of the genome
    starts = starts[ends < len(snps)]
    ends = ends[ends < len(snps)]
    ends = ends[starts >= 0]
    starts = starts[starts >= 0]

    print(len(starts), 'windows with GWAS hits found')

    # compute correlations
    numbers = pd.DataFrame(
            np.array([[i,j,
                np.max(snps.iloc[i:j].chi2),
                np.corrcoef(snps.iloc[i:j].Rv_resid, snps.iloc[i:j].ahat_resid)[0,1]]
                for i,j in zip(starts, ends)]),
            columns=['start','end','maxchi2','corr'])

    # keep only cases with strong correlations in the right direction
    numbers = numbers[numbers['corr']**2 >= corr_thresh]
    numbers = numbers[np.sign(numbers['corr']) == np.sign(z)]
    print(len(numbers), 'windows with GWAS hits and squared correlation with Rv >=',
            corr_thresh, 'in the right direction')
    for i,j in zip(numbers.start, numbers.end):
        i = int(i); j = int(j)
        print('saving', i,j)
        c = snps.iloc[i].CHR
        start = snps.iloc[i].BP
        end = snps.iloc[j].BP
        plt.figure()
        plt.scatter(snps.iloc[i:j].Rv_resid,
                snps.iloc[i:j].ahat_resid * np.sqrt(snps.iloc[i:j].N))
        plt.title('chr{}:{}-{}'.format(c, start, end))
        plt.xlabel(r'residual $Rv$')
        plt.ylabel(r'residual $Z$')

        filename = '{}/chr{}:{}-{}.pdf'.format(folder, c, start, end)
        fs.makedir_for_file(filename)
        plt.savefig(filename)
        plt.close()
